#!/usr/bin/python
# -*- coding: utf-8 -*-

# csv_file.py
# 2007 Sep 20 . ccr

"""Read/write comma-separated-value (*.csv) files.

"""

# ccr . 2011 Jan 22 . Fix encode None.
# ccr . 2011 Jan 19 . Fix Unicode I/O.
# ccr . 2010 Apr 02 . Refactor to expose iterators.

from __future__ import division
import sys
import os
import codecs
import cStringIO  # 2001 Jan 19
import csv as csv_module  # 2001 Jan 19

ZERO = 0
SPACE = ' '
NULL = ''
NUL = '\x00'
NA = -1
ASCII_SAFE = 'unicode-escape'


class UTF8Recoder:  # 2001 Jan 19

    """Iterator that reads an encoded stream and reencodes the input to UTF-8.

    This class is borrowed from an example in the documentation for
    the standard Python 2.7 distribution.
    
    """
    
    def __init__(self, f, encoding):
        self.reader = codecs.getreader(encoding)(f)
        return

    def __iter__(self):
        return self

    def next(self):
        return self.reader.next().encode("utf-8")


class UnicodeReader:  # 2001 Jan 19

    """A CSV reader which will iterate over lines in the CSV file 'f',
    which is encoded in the given encoding.
    
    This class is borrowed from an example in the documentation for
    the standard Python 2.7 distribution.
    
    """

    def __init__(self, f, dialect=csv_module.excel, encoding="utf-8", **kwds):
        f = UTF8Recoder(f, encoding)
        self.reader = csv_module.reader(f, dialect=dialect, **kwds)
        self.line_num = ZERO
        return

    def next(self):
        row = self.reader.next()
        return [unicode(s, "utf-8") for s in row]

    def __iter__(self):
        return self


class UnicodeWriter:  # 2001 Jan 19

    """A CSV writer which will write rows to CSV file 'f', which is
    encoded in the given encoding.
    
    This class is borrowed from an example in the documentation for
    the standard Python 2.7 distribution.
    
    """

    def __init__(self, f, dialect=csv_module.excel, encoding="utf-8", **kwds):
        # Redirect output to a queue
        self.queue = cStringIO.StringIO()
        self.writer = csv_module.writer(self.queue, dialect=dialect, **kwds)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()
        return

    def writerow(self, row, missing=None):
        ascii_safe = []
        for s in row:
            if s is None:
                ascii_safe.append(s)  # 2011 Jan 22
            else:
                ascii_safe.append(s.encode("utf-8"))
        self.writer.writerow(ascii_safe)
        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        data = data.decode("utf-8")
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)
        return self

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)
        return self
    

class DictReader(csv_module.DictReader):  # 2011 Jan 19

    """Subclass the csv.DictReader in the standard library to
    substitute UnicodeReader for the default reader.

    """

    def __init__(
        self,
        f,
        fieldnames=None,
        restkey=None,
        restval=None,
        dialect="excel",
        reader=UnicodeReader,
        *args,
        **kwds
        ):
        csv_module.DictReader.__init__(
            self,
            f=f,
            fieldnames=fieldnames,
            restkey=restkey,
            restval=restval,
            dialect=dialect,
            )
        self.reader = reader(f, dialect, *args, **kwds)
        return


class DictWriter(csv_module.DictWriter):  # 2011 Jan 19

    """Subclass the csv.DictWriter in the standard library to
    substitute UnicodeWriter for the default writer.

    """

    def __init__(
        self,
        f,
        fieldnames,
        restval="",
        extrasaction="raise",
        dialect="excel",
        writer=UnicodeWriter,
        *args,
        **kwds
        ):
        csv_module.DictWriter.__init__(
            self,
            f=f,
            fieldnames=fieldnames,
            restval=restval,
            extrasaction=extrasaction,
            dialect=dialect,
            )
        self.writer = writer(f, dialect, *args, **kwds)
        return

        
class CsvFile(list):

    '''Container for file records.

    This list is populated by the *load* method, which brings all the
    records of the whole file into memory at once.  Alternatively, you
    may call the *iter_record* method, which does not populate the
    list but one-at-a-time yields only the next record.

    This class encapsulates all I/O for *.csv files.

    '''

    def __init__(self, tags=None, encoding='utf-8'):

        '''Instantiate a *.csv file.

        You may read or write files.

        A list of tags (field names) must be provided for files you
        want to write.  Column order of fields in the header and in
        each record (that is, on each line) is the same as the order
        of the tags in this list.

        '''
        
        list.__init__(self)
        if tags is None:
            self.tags = []
        else:
            self.tags = tags
        self.encoding = encoding
        return

    def eval_row(self, rec):

        '''Unescape newlines in input data.'''
        
        result = {}
        for (key, value) in rec.iteritems():
            if value is None:
                result[key] = None
            else:
                result[key] = value.replace(r'\n', '\n')
        return result

    def repr_row(self, rec):

        '''Escape newlines in output data.'''
        
        result = {}
        for (key, value) in rec.iteritems():
            if value is None:
                result[key] = None
            else:
                result[key] = value.replace('\n', r'\n')
        return result

    def load(self, flob=None, progress=False):

        '''Read all records (lines) in a file.

        An instance of this class is a list of records, which is empty
        until after the load is complete.

        A record is a dictionary of values indexed by field name.

        You may load from any file.  Flob is any file-like object or a
        file name.

        If progress is True, an ASCII-art bar graphic is displayed on
        the standard error device during loading.

        '''
        
        self.init_reader(flob)
        if progress:
            self.progress_init()
        for row in self.iter_record():
            self.append(row)
            if progress:
                self.progress_cont()
        if progress:
            self.progress_term()
        self.close()
        return self

    def save(self, flob=None):

        '''Write all records appended to this class instance.

        A record is a dictionary of values indexed by field name.

        You may write to any file.  Flob is any file-like object or a
        file name.

        '''
        
        self.init_writer(flob)
        for row in self:
            self.put_record(row)
        self.close()
        return self

    def close(self):

        '''Wrap up I/O to flob.

        You should call this after I/O is done.

        '''
        
        if self.needs_close:
            self.flob.close()
        return self

    def progress_init(self):

        '''Initialize progress bar graphic.

        This is called automagically by the load method.  You may wish
        to call it yourself if you call the iter_record method.

        '''
        
        self.progress_count = ZERO
        self.progress = ZERO
        self.progress_inc = 1
        return self

    def progress_cont(self):

        '''Increment counts for progress bar graphic.

        This is called automagically by the load method.  You may wish
        to call it yourself if you call the iter_record method.

        '''
        
        self.progress_count += 1
        if (self.progress_count % 100) == ZERO:
            self.progress_show()
            self.progress = self.progress + self.progress_inc
            if self.progress >= 50:
                self.progress = 50
                self.progress_inc = -1
            elif self.progress <= ZERO:
                self.progress = ZERO
                self.progress_inc = 1
        return self

    def progress_show(self):

        '''Display progress bar graphic.

        This is called automagically.

        '''
        
        prolog = SPACE * self.progress
        epilog = SPACE * (50 - self.progress)
        sys.stderr.write('\r[%s<=>%s] %5i records.' % (prolog, epilog, self.progress_count))
        return self

    def progress_term(self):

        '''Wrap up use of the progress bar graphic.

        This is called automagically by the load method.  You may wish
        to call it yourself if you call the iter_record method.

        '''
        
        self.progress_show()
        sys.stderr.write('\n')
        return self

    def iter_record(self):

        '''Iterate over records in flob.

        You may use this method to read each record in sequence
        without loading all records all at once.  Some files are just
        too big to do that.

        A record is a dictionary of values indexed by field name.

        '''
        
        for rec in self.reader:
            yield self.eval_row(rec)
        return

    def put_record(self, rec):

        '''Write a single record to flob.

        You may use this method to place records in the output file
        one at a time.

        A record is a dictionary of values indexed by field name.

        '''
        
        self.writer.writerow(self.repr_row(rec))
        return self

    def init_reader(self, flob, restval=None, restkey=None):

        '''Set up input file.

        Flob is any file-like object or a file name.

        Do this before calling the iter_record method.

        Do not do this before calling the load method.

        '''

        self.flob = self.get_input_unit(flob)
        queue = cStringIO.StringIO()  # 2001 Jan 19
        queue.write(self.flob.readline())  # 2001 Jan 19
        queue.seek(ZERO)  # 2001 Jan 19
        header_reader = UnicodeReader(queue)  # 2001 Jan 19
        self.tags = header_reader.next()
        self.reader = DictReader(
            self.flob,
            self.tags,
            restval=restval,
            restkey=restkey,
            encoding=self.encoding,
            )  # 2001 Jan 19
        return self

    def init_writer(self, flob, restval=NULL):

        '''Set up output file.

        Flob is any file-like object or a file name.

        Do this before calling the put_record method.

        Do not do this before calling the save method.

        '''
        
        self.flob = self.get_output_unit(flob)
        self.writer = DictWriter(
            self.flob,
            self.tags,
            restval=restval,
            encoding=self.encoding,
            )  # 2001 Jan 19
        header = dict([(col, col) for col in self.tags])
        self.put_record(header)
        return self

    def get_input_unit(self, flob):

        '''Return an open file-like object for reading.

        This is called automagically.

        '''
        
        if flob in [None, '-']:
            result = sys.stdin
            self.needs_close = False
        elif hasattr(flob, 'read'):
            result = flob
            self.needs_close = False
        else:
            flob = os.path.expanduser(flob)
            result = open(flob, 'rb')
            self.needs_close = True
        return result

    def get_output_unit(self, flob):

        '''Return an open file-like object for writing.

        This is called automagically.

        '''
        
        if flob in [None, '-']:
            result = sys.stdout
            self.needs_close = False
        elif hasattr(flob, 'write'):
            result = flob
            self.needs_close = False
        else:
            flob = os.path.expanduser(flob)
            result = open(flob, 'wb')
            self.needs_close = True
        return result


def main_line():

    r"""Round-trip testing.

    >>> text = 'animal, count \ncat, 2 \ndog, 3'
    >>> import StringIO
    >>> unit = StringIO.StringIO(text)
    >>> csv = CsvFile()
    >>> csv.load(unit)
    [{u' count ': u' 2 ', u'animal': u'cat'}, {u' count ': u' 3', u'animal': u'dog'}]
    >>> unit = StringIO.StringIO()
    >>> csv.save(unit)
    [{u' count ': u' 2 ', u'animal': u'cat'}, {u' count ': u' 3', u'animal': u'dog'}]
    >>> unit.getvalue()
    'animal, count \r\ncat, 2 \r\ndog, 3\r\n'
    """

    import doctest
    doctest.testmod()
    stdout = codecs.getwriter('utf-8')(sys.stdout)
    csv = CsvFile(tags=['payee', 'balance'])
    csv.append({'payee':u'Acme\N{TRADE MARK SIGN}', 'balance':repr(12.34)})
    stdout.write('%s\n' % str(csv))
    stdout.write('%s\n' % csv[ZERO]['payee'])
    csv.save('/tmp/test.csv')
    unit = codecs.open('/tmp/test.csv', 'r', 'utf-8')
    stdout.write(unit.read())
    unit.close()
    csv=CsvFile()
    csv.load('/tmp/test.csv')
    stdout.write('%s\n' % str(csv))
    stdout.write('%s\n' % csv[ZERO]['payee'])
    return


if __name__ == "__main__":
    main_line()


# Fin
