#!/usr/bin/python
# -*- coding: utf-8 -*-

# odf.py
# 2010 Jan 22 . ccr

"""Read and update (but not create) Open Document Format Text files.

This module contains building blocks that Python scripts can use to
generate well-formatted reports.  The reports may then be convered to
*.pdf files for distribution using OpenOffice.org.

For navigating XML, this script uses *lxml*, which is not in the
standard Python distribution.  It is available as a Debian package.

"""

# 2011 Jan 20 . ccr . Refactor.  Fix encoding.
# 2010 Dec 06 . ccr . Insert doc strings.

from __future__ import division
import sys
import os
import re
import time
import signal
import subprocess
import zipfile
import tempfile
import lxml.etree as ET

ZERO = 0
SPACE = ' '
NULL = ''
NUL = '\x00'
NA = -1

COMMAND_PIPE = subprocess.PIPE
COMMAND_STDOUT = subprocess.STDOUT
PAT_TAG_CLOSE = re.compile(r'<([^<]*?)((?:\s+[^<]+)?)\s*/>')
URI = {
    "office": "urn:oasis:names:tc:opendocument:xmlns:office:1.0",
    "style": "urn:oasis:names:tc:opendocument:xmlns:style:1.0",
    "text": "urn:oasis:names:tc:opendocument:xmlns:text:1.0",
    "table": "urn:oasis:names:tc:opendocument:xmlns:table:1.0",
    "draw": "urn:oasis:names:tc:opendocument:xmlns:drawing:1.0",
    "fo": "urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0",
    "xlink": "http://www.w3.org/1999/xlink",
    "dc": "http://purl.org/dc/elements/1.1/",
    "meta": "urn:oasis:names:tc:opendocument:xmlns:meta:1.0",
    "number": "urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0",
    "svg": "urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0",
    "chart": "urn:oasis:names:tc:opendocument:xmlns:chart:1.0",
    "dr3d": "urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0",
    "math": "http://www.w3.org/1998/Math/MathML",
    "form": "urn:oasis:names:tc:opendocument:xmlns:form:1.0",
    "script": "urn:oasis:names:tc:opendocument:xmlns:script:1.0",
    "ooo": "http://openoffice.org/2004/office",
    "ooow": "http://openoffice.org/2004/writer",
    "oooc": "http://openoffice.org/2004/calc",
    "dom": "http://www.w3.org/2001/xml-events",
    "xforms": "http://www.w3.org/2002/xforms",
    "xsd": "http://www.w3.org/2001/XMLSchema",
    "xsi": "http://www.w3.org/2001/XMLSchema-instance",
    "field": "urn:openoffice:names:experimental:ooxml-odf-interop:xmlns:field:1.0",
    }


def ns(path):
    
    """Expand namespace prefixes.
    
    """
        
    tags = path.split('/')
    for (ndx, tag) in enumerate(tags):
        qfn = tag.split(':')
        if len(qfn) > 1:
            (prefix, tag) = qfn[:2]
            prefix = URI[prefix]
            tag = '{%s}%s' % (prefix, tag)
            tags[ndx] = tag
    path = '/'.join(tags)
    return path
    

def get_text(soup):

    """Return space-delimited text from an XML tag (and its children).

    """
    
    return ET.tostring(soup, method='text')


def protect_ampersand(text):

    """Change text to XML entities.

    """
    
    result = text.replace('&', '&amp;')
    return result


def protect_lt_gt(text):

    """Change text to XML entities.

    """
    
    result = text.replace('<', '&lt;').replace('>', '&gt;')
    return result


class Template(object):

    """XML tag template.

    Find a qualifying tag in parent, yank it, and insert clones in its
    place.

    """

    def extract(self, elt):
        if elt is None:
            self.parent = None
            self.origin = None
            self.text = None
        else:
            self.parent = elt.getparent()
            self.origin = self.parent.index(elt)
            self.parent.remove(elt)
            self.text = ET.tostring(elt)
        self.ndx = self.origin
        return self

    def reset(self, parent):
        self.parent = parent
        self.ndx = self.origin
        return self

    def insert_in_parent(self, targets=None):
        if targets is None:
            targets = []
        text = self.text
        for (old, new) in targets:
            if new is None:
                new = NULL
            text = text.replace(old, new)
        result = ET.fromstring(text)
        self.parent.insert(self.ndx, result)
        self.ndx += 1
        return result


class Archive(object):

    """An Open Document Format Text file.

    This is a zip-compressed archive.

    """    

    def load(self, flob):

        """Open a file-like object as an ODT file archive.

        """
        
        self.flob = flob
        doc = zipfile.ZipFile(file=self.flob, mode='r')
        self.members = dict([
            (zipinfo.filename, [zipinfo, doc.read(zipinfo.filename)])
            for zipinfo in doc.infolist()
            ])
        doc.close()
        return self

    def save(self, flob=None):

        """Replace all members in an ODT file archive.

        """
        
        if flob is None:
            flob = self.flob
        doc = zipfile.ZipFile(file=flob, mode='w')
        for (member, [zipinfo, content]) in self.members.iteritems():
            doc.writestr(zipinfo, content)
        doc.close()
        return self

    def update(self, member, text):

        """Replace one member in an ODT file archive.

        This is deprecated because it cannot overwrite the old member.
        Instead it merely appends a substitute, which -- in the case
        of content -- practically doubles the size of the archive.

        """
        
        self.set_text(self, member, text)
        doc = zipfile.ZipFile(file=self.flob, mode='a')
        (zipinfo, content) = self.members[member]
        doc.writestr(zipinfo, content)
        doc.close()
        return self

    def get_text(self, member):
        return self.members[member][1]

    def set_text(self, member, text):
        self.members[member][1] = text
        return self

    def get_soup(self, member):

        """Return a structure representing the XML
        content.

        """
        
        result = ET.fromstring(self.get_text(member))
        return result

    def set_soup(self, member, soup):

        """Set member content to an XML string.

        """
        
        self.members[member][1] = str(soup)
        return self
    

class Content(Archive):

    """Helper object to manipulate the 'content.xml' member of an ODF
    file.

    """    

    def load(self, flob):
        Archive.load(self, flob=flob)
        return self.get_text('content.xml').decode('utf-8')  # 2011 Jan 20

    def save(self, text, flob=None):
        self.set_text('content.xml', text.encode('utf-8'))  # 2011 Jan 20
        Archive.save(self, flob=flob)
        return self
        
    def load_soup(self, flob):
        Archive.load(self, flob=flob)
        return self.get_soup('content.xml')

    def save_soup(self, soup, flob=None):
        self.set_soup('content.xml', soup)
        Archive.save(self, flob=flob)
        return self
        
    def update(self, text):
        Archive.update(self, 'content.xml', text)
        return self


class TempPath(object):

    """Draw a random name for a temporary directory.

    """

    def __init__(self, prefix=NULL):
        self.name = tempfile.mkdtemp(prefix=prefix)
        return

    def remove(self):
        for (root, dirs, files) in os.walk(self.name, topdown=False):
            for name in files:
                os.remove(os.path.join(root, name))
            for name in dirs:
                os.rmdir(os.path.join(root, name))
        os.rmdir(self.name)
        return self


class Command(subprocess.Popen):

    """Run shell command.

    """

    def __init__(self,
                 args,
                 stdin=None,
                 stdout=None,
                 stderr=None,
                 close_fds=False,
                 shell=True,
                 cwd=None,
                 env=None,
                 ):

        subprocess.Popen.__init__(self,
                         args=args,
                         stdin=stdin,
                         stdout=stdout,
                         stderr=stderr,
                         close_fds=close_fds,
                         shell=shell,
                         cwd=cwd,
                         env=env,
                         )
        return


class SofficeListener(Command):

    """Create a listener deamon for OpenOffice.org to handle
    automated file-conversion requests.

    *unoconv* 0.3-3 is available as a Debian Linux 5.0 (Lenny)
    package.

    """

    def __init__(self):
        try:
            os.remove('__db.')  # Ugly workaround for Debian Bug $505126 against *unoconv*.
        except OSError:
            pass
        Command.__init__(
            self,
            args='unoconv --listener',
            stdout=COMMAND_PIPE,
            stderr=COMMAND_STDOUT,
            )
        assert self.is_active()
        time.sleep(30)
        return

    def is_active(self):
        count = 10
        while True:
            status = Command(args='pgrep soffice', stdout=COMMAND_PIPE)
            status.wait()
            soffice_pid = status.stdout.read().strip()
            try:
                soffice_pid = int(soffice_pid, 10)
                result = soffice_pid == self.pid
            except ValueError:
                result = False
            if result:
                break
            count -= 1
            if count is ZERO:
                break
            time.sleep(1)
        return result

    def shut_down(self):
        assert self.is_active()
        os.kill(self.pid, signal.SIGTERM)
        self.wait()
        return self


class SofficeConverter(Command):

    """Convert an OpenOffice.org ODF file.

    *unoconv* 0.3-3 is available as a Debian Linux 5.0 (Lenny)
    package.

    """

    def __init__(self, input_doc_name, output_format='pdf'):
        Command.__init__(
            self,
            args='unoconv --format=%s "%s"' % (output_format, input_doc_name),
            stdout=COMMAND_PIPE,
            stderr=COMMAND_STDOUT,
            )
        return


class ConcatenatePdf(Command):

    """Concatenate a series of *.pdf files.

    *pdftk* 1.41-3 is available as a Debian Linux 5.0 (Lenny) package.

    """

    def __init__(self, input_doc_names, output_doc_name):
        Command.__init__(
            self,
            args='pdftk "%s" cat output "%s"' % (
                '" "'.join(input_doc_names),
                output_doc_name,
                ),
            )
        return


class SpoolPdf(Command):

    """Send a (*.pdf) document to a printer queue using the standard
    Linux print utility.

    """

    def __init__(self, output_doc_name, queue=None):
        if queue is None:
            lp_command = 'lp "%s"' % output_doc_name
        else:
            lp_command = 'lp -d "%s" "%s"' % (queue, output_doc_name)
        Command.__init__(
            self,
            args=lp_command,
            stdout=COMMAND_PIPE,
            stderr=COMMAND_STDOUT,
            )
        return


def main_line():

    """Testing code goes here.

    """
    
    return


if __name__ == "__main__":
    main_line()


# Fin
