#!/usr/bin/python2.5
# -*- coding: utf-8 -*-

"""pretty_checks/pcs.py
Written:  2011 Jan 21

Prettify the printing of check forms for GnuCash users.

Copyright (C) 2011  Charles Curtis Rhode

    This program is free software: You can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see
    <http://www.gnu.org/licenses/>.

This package of Python scripts depends on OpenOffice as well as
GnuCash.

GnuCash is a 'personal finance tracking program.'  It is available as
a Debian Linux package.

OpenOffice is a 'full-featured office productivity suite.'  It, too,
is available as a Debian Linux package.

The check writing functionality of GnuCash is rudimentary at version
2.2+.  Pretty Checks reads and updates GnuCash books while creating
check documents that can be printed with all the specialized font and
text formatting features of OpenOffice.  Pretty Checks has been used
on preprinted cut-sheet safety-paper personal check blanks purchased
from Walmart Financial Services at <http://www.walmartchecks.com>.

The only other thing (besides Python) that Pretty Checks needs is
lxml.etree, which supports manipulation of xml files.  This is
available as a Debian Linux package, as well.

The scripts in this package update GnuCash books.  The techniques they
use can be a source of error.  Obviously any problems they create with
GnuCash are the fault of Pretty Checks and not of GnuCash.  It should
go without saying that you should log out of GnuCash and stay out
while running Pretty Checks scripts.

All script names for Pretty Checks have symbolic links to this script,
pcs.py.  The parameters each requires are documented by invoking each
script with the --help option.  The scripts are:

"""

from __future__ import division

DOC_STRINGS = {}

DOC_STRINGS['pcs'] = """

Do initial setup of Pretty Checks.  This script is normally used once.

As you may by now have surmised, you can unpack the Pretty Checks
archive like this:

 > tar -xvf pretty_checks.tar.gz

Then you should probably do:

 > cd pretty_checks
 > python pcs.py --help

Then you should definitely do:

 > python pcs.py --setup

... right away.

You are now ready to start printing Pretty Checks.

"""

DOC_STRINGS['pcs_extract_vendors'] = """

Open a GnuCash book to retrieve the list of vendors and write a
comma-separated-value text database for merging with other address
databases.

The name of the csv file may be provided.  If omitted, the result is
printed on standard output.

This script is not normally used.  It is provided for applications
other than printing checks.

"""

DOC_STRINGS['pcs_merge_vendors'] = """

Merge a list of vendors in comma-separated-value format with a
GnuCash book.

The name of the csv file may be provided.  If omitted, the vendors
list is read from standard input.

This script is dangerous because it can change vendor names on past
transactions.  It should not normally be required.

"""

DOC_STRINGS['pcs_extract_payments'] = """

Open a GnuCash book to retrieve credits against bank accounts.

A bank account name is required.  Only transactions in the given
account that have an asterisk (*) in the form number are selected.

Output is a comma-separated-value file.  The name of the csv file may
be provided.  If omitted, the result is written on standard output.
The result may be edited if desired but need not be.

When printed via the reprint checks script, these transactions become
checks and are numbered.  Then they should be merged back into the
GnuCash book with the merge payments script to preserve the record of
their form numbers.

There is a sleight here: Each transaction description is deemed to be
the name of the payee.  However, if it matches the company name of a
GnuCash vendor, the vendor's address is available for printing on the
check stub.

In Pretty Checks' scheme of things, the company name of a GnuCash
vendor, the vendor's payment address name, and the first line of the
vendor's payment address would generally all be the same.  I suppose
there are shady exceptions where the check could be made out in a
person's name but mailed to a post office box.  Probably it's not
particularly unusual for payees to specify their checks be mailed to
agents.  In such cases the check is made out to a payee (vendor's
payment address name) whose name need not appear anywhere in the
actual lines of the vendor's payment address.  Note that the
transaction description, which should point to the company name of a
GnuCash vendor, is not necessarily the same as the payee (vendor's
payment address name).  It is, of course, the same when there is no
matching vendor in the Gnucash book.

"""

SKELETON_FIELDS = [
    'date_spelled',
    'date_mm_dd_yyyy',
    'date_mm',
    'date_dd',
    'date_yyyy',
    'form_num',
    'dollar_amt',
    'dollars',
    'cents',
    'dollars_spelled',
    'memo',
    'vendor_name',
    'vendor_id',
    'payee',
    'addr1',
    'addr2',
    'addr3',
    'addr4',
    'phone',
    'fax',
    'email',
    ]

DOC_STRINGS['pcs_reprint_checks'] = """

Update the comma-separated-value file of check transactions to add
form numbers.

The name of the csv file may be provided.  If provided, it is first
read and then overwritten.  If omitted, the transactions are first
read from standard input and then copied with form numbers to standard
output.

This script uses the GnuCash book to retrieve vendor info but does
not update the GnuCash book.

The name of an Open Document Format text file may be provided, which
contains a check-form skeleton.  This skeleton may (or may not)
contain any of the following strings, which are expanded with actual
values for each check to be printed:

 <%s>

All the text should be in a document 'section.'  The first paragraph
should force a prior page break in its text-flow property.  The first
section in the skeleton will be duplicated for each check, and thus
each check will start on a new page.

A name for an Open Document Format text output file must be
provided.  This will contain all the check forms with page breaks.
Open it in OpenOffice and print the checks from there.

Each run starts with a beginning form number and assigns consecutive
numbers to succeeding forms.  These numbers must match the ones on
the prenumbered forms so that checks may later be reconciled against
a bank statement.

It is good practice while designing the layout of the check form
skeleton to place the generated number next to the preprinted one on
each stub for convenient visual comparison at the end of the run.

There is no 'print checks' script.  Simply run the reprint checks
script to generate the forms, and rerun it as necessary to overcome
printer jams, recipient, amount and issue-date mistakes, and other
user errors.  It will be more straightforward to destroy all the check
forms from an aborted run and start over with new numbers rather than
void some forms and not others.

When satisfactory, the assigned form numbers should be merged back
into the GnuCash book with the merge payments script.

""" % """>,
 <""".join(SKELETON_FIELDS)

DOC_STRINGS['pcs_void_checks'] = """

Update the GnuCash book to insert void transactions to record the
sequence numbers of the forms scratched in attempts to run printed
checks.

Accountants and auditors like to keep track of missing numbered forms.

A bank account name is required.

"""

DOC_STRINGS['pcs_merge_payments'] = """

Record actual form numbers in the GnuCash book.

This script does what the extract payments and the reprint checks
scripts do not.  Also, it makes the transaction dates current, and it
records the actual amount from the face of the check in case that was
changed in the csv file from what it was in the GnuCash book, creating
an out-of-balance condition.

The name of the csv file may be provided.  If omitted, the
transactions are read from standard input.

This script should be run after the check forms are processed, signed,
and ready for mailing.

"""

# Here is an ordered list of script names as they appear in
# documentation:

SCRIPT_NAMES = [
    'pcs_extract_vendors',
    'pcs_merge_vendors',
    'pcs_extract_payments',
    'pcs_reprint_checks',
    'pcs_void_checks',
    'pcs_merge_payments',
    ]

# ccr . V1.1 . 2001 Jan 21 . First production.

# Pretty Checks is primarily a word-processing merge application, so
# the check-form skeleton is an Open Document Format text file , but,
# because the OpenOffice.org (OO.org) merge function is so opaque, I
# resort to manipulating the odt text skeleton directly and achieve the
# same effect without running an actual document merge.
# 
# All the scripts are in this one big script.  The only externals not
# supplied by the standard Python distribution are:
#
# o lxml.etree, which supports manipulation of xml files.  This is
# available as a Debian package.  The modules in the standard Python
# distribution: xml.dom, xml.etree, sax, et al are not sufficient.
# Neither is BeautifulSoup.  Standard distribution xml.etree
# serialization is inadequate.
#
# o odf_file.py, which supports manipulation of the Open Document
# Format text check-form skeleton file.  This is packaged with Pretty
# Checks.
#
# o csv_file.py, which supports manipulation of comma-separated-values
# text files.  This is packaged with Pretty Checks.
#
# The objects that support manipulation of the GnuCash book descend
# from GncObject in this script and could just as well go into a
# separate module if it were necessary to support separate scripts.
# At this point that is not convenient.
#
# I've decided to put off designing a graphical user interface.  A GUI
# can always be wrapped around the scripts.

import sys
import os
import codecs
import optparse
import lxml.etree as ET
import re
import gzip
import datetime
import csv_file
import odf_file

ZERO = 0
SPACE = ' '
NULL = ''
NUL = '\x00'
NA = -1
EMPTY = [None, NULL]

DEBUG = False
CODING = 'utf-8'
PCS_VERSION = '1.1'  # 2011 Jan 21
GNC_VERSION = '2.0.0'  # The compatible GnuCash version(s) are
                       # asserted.
CURRENCY_WORDS = {
    ZERO: "Zero",
    1: "One",
    2: "Two",
    3: "Three",
    4: "Four",
    5: "Five",
    6: "Six",
    7: "Seven",
    8: "Eight",
    9: "Nine",
    10: "Ten",
    11: "Eleven",
    12: "Twelve",
    13: "Thirteen",
    14: "Fourteen",
    15: "Fifteen",
    16: "Sixteen",
    17: "Seventeen",
    18: "Eightteen",
    19: "Nineteen",
    20: "Twenty",
    30: "Thirty",
    40: "Forty",
    50: "Fifty",
    60: "Sixty",
    70: "Seventy",
    80: "Eighty",
    90: "Ninety",
    }
PAT_NUMERIC = re.compile(r'(\d+)')
PAT_GNC_DATE = re.compile(r'^(\d{4})-(\d{2})-(\d{2}) '
                          r'(\d{2}):(\d{2}):(\d{2})(?:\.\d+)? '
                          r'([+-]\d{2})(\d{2})$')


class StdWriter(object):

    """Implement character encoding on standard output.

    """

    def __init__(self, flob=sys.stdout):
        self.stdout = codecs.getwriter(CODING)(flob, errors='replace')
        return

    def put(self, text):
        self.stdout.write(text)
        return self


class StdReader(object):

    """Implement character encoding on standard input.

    """

    def __init__(self, flob=sys.stdin):
        self.stdin = codecs.getreader(CODING)(flob, errors='replace')
        return

    def get(self):
        return self.stdin.readline()


STD_READER = StdReader()
STD_WRITER = StdWriter()
STD_ERROR = StdWriter(flob=sys.stderr)


class GncDateTime(datetime.datetime):

    """A subclass of standard distribution datetime.datetime.

    This object prints like a GnuCash xml date object.  Also it has
    its own *repr* representation for csv serialization.

    """

    def __str__(self):
        (delta_hrs, delta_min) = divmod(self.delta_min, 60)
        result = self.strftime('%%Y-%%m-%%d %%H:%%M:%%S %+03i%02i' % (
            delta_hrs,
            delta_min,
            ))
        return result

    def set_delta(self, delta_min):
        self.delta_min = delta_min
        return self

    def __repr__(self):
        (year, month, day, hours, minutes, seconds) = self.timetuple()[:6]
        result = 'GncDateTime(%i, %i, %i, %i, %i, %i).set_delta(%i)' % (
            year,
            month,
            day,
            hours,
            minutes,
            seconds,
            self.delta_min,
            )
        return result

NOW = GncDateTime.utcnow()
TZ_OFFSET = (datetime.datetime.now() - NOW)
TZ_OFFSET_MINUTES = TZ_OFFSET.days * 24 * 60 + round(TZ_OFFSET.seconds / 60)
NOW.set_delta(TZ_OFFSET_MINUTES)
TODAY = GncDateTime.combine(NOW, datetime.time(ZERO)).set_delta(TZ_OFFSET_MINUTES)

def Spell(dollars):

    """Convert currency to words.

    >>> Spell(42341.0)
    'Forty-Two Thousand Three Hundred Forty-One'
    >>> Spell(ZERO)
    'No'
    >>> Spell(1)
    'One'
    >>> Spell(76842341)
    'Seventy-Six Million Eight Hundred Forty-Two Thousand Three Hundred Forty-One'
    >>> Spell(11)
    'Eleven'
    >>> Spell(456)
    'Four Hundred Fifty-Six'
    >>> Spell(76)
    'Seventy-Six'
    >>> Spell(777777777)
    'Seven Hundred Seventy-Seven Million Seven Hundred Seventy-Seven Thousand Seven Hundred Seventy-Seven'
    >>> Spell(800)
    'Eight Hundred'
    >>> Spell(2000)
    'Two Thousand'
    >>> Spell(40)
    'Forty'
    """
    
    def hundreds(dollars):
        result = []
        (hundreds, dollars) = divmod(dollars, 100)
        if hundreds > ZERO:
            result.append(CURRENCY_WORDS[hundreds])
            result.append('Hundred')
        if dollars == ZERO:
            pass
        elif (dollars < 20):
            result.append(CURRENCY_WORDS[dollars])
        else:
            (tens, dollars) = divmod(dollars, 10)
            if dollars == ZERO:
                result.append('%s' % CURRENCY_WORDS[tens * 10])
            else:
                result.append('%s-%s' % (CURRENCY_WORDS[tens * 10], CURRENCY_WORDS[dollars]))
        return result

    result = []
    if dollars == ZERO:
        result.append('No')
    else:
        (millions, dollars) = divmod(dollars, 10**6)
        if millions > ZERO:
            result.extend(hundreds(millions))
            result.append('Million')
        (thousands, dollars) = divmod(dollars, 1000)
        if thousands > ZERO:
            result.extend(hundreds(thousands))
            result.append('Thousand')
        result.extend(hundreds(dollars))        
    return SPACE.join(result)


class Check(object):

    """Check form.

    """

    def from_gnc(self, transaction, split, vendor=None):
        self.transaction = transaction
        self.form_num = self.transaction.form_num
        self.date_posted = self.transaction.date_posted
        self.split = split
        self.value = - self.split.value
        self.memo = self.split.memo
        self.split_uid = self.split.uid
        self.vendor_name = self.transaction.description
        self.from_vendor(vendor)
        return self

    def from_vendor(self, vendor):
        self.vendor = vendor
        if self.vendor is None:
            self.vendor_id = None
            self.vendor_address = None
            self.payee = self.vendor_name
            self.addr1 = None
            self.addr2 = None
            self.addr3 = None
            self.addr4 = None
            self.phone = None
            self.fax = None
            self.email = None
        else:
            self.vendor_id = self.vendor.id
            self.vendor_address = self.vendor.address
            self.payee = self.vendor_address.addr_name
            self.addr1 = self.vendor_address.addr_1
            self.addr2 = self.vendor_address.addr_2
            self.addr3 = self.vendor_address.addr_3
            self.addr4 = self.vendor_address.addr_4
            self.phone = self.vendor_address.phone
            self.fax = self.vendor_address.fax
            self.email = self.vendor_address.email
        return self

    def to_odf(self, page):
        self.date_spelled = self.date_posted.strftime('%a %d %b %Y')
        self.date_mm_dd_yyyy = self.date_posted.strftime('%m/%d/%Y')
        (self.date_mm, self.date_dd, self.date_yyyy) = self.date_mm_dd_yyyy.split('/')
        self.dollar_amt = '%.2f' % self.value
        (self.dollars, self.cents) = divmod(self.value * 100.0, 100.0)
        self.dollars_spelled = Spell(self.dollars).capitalize()
        (self.dollars, self.cents) = ('%i' % self.dollars, '%02i' % self.cents)
        targets = [('&lt;%s&gt;' % tag, getattr(self, tag, NULL))
            for tag in SKELETON_FIELDS
            ]
        page.insert_in_parent(targets)
        return self

    def to_gnc(self, elt_split):
        split = GncObject().from_elt(elt_split)
        split.put_dollars('split:value', - self.value)
        split.put('split:action', 'check', can_add=True)
        elt_transaction = elt_split.getparent().getparent()
        transaction = GncObject().from_elt(elt_transaction)
        transaction.put('trn:num', self.form_num)
        transaction.put_date('trn:date-posted', self.date_posted)
        return self


class Csv(csv_file.CsvFile):

    """Abstract csv file.

    """

    tags = []

    def __init__(self):
        csv_file.CsvFile.__init__(self, tags=self.tags)
        return


class CsvVendors(Csv):

    """List tags used on csv vendor files.

    """

    tags_id = [
        'uid',
        'name',
        'id',
        'notes',
        ]
    tags_addr = [
        'addr_name',
        'addr_1',
        'addr_2',
        'addr_3',
        'addr_4',
        'phone',
        'fax',
        'email',
        ]
    tags = tags_id + tags_addr

    pass


class CsvPayments(Csv):

    """List tags used on csv payments files.

    """

    tag_functions = [
        ('form_num', None, None),
        ('split_uid', None, None),
        ('date_posted', repr, eval),
        ('value', '%.2f', float),
        ('vendor_name', None, None),
        ('memo', None, None),
        ]
    tags = [tag for (tag, f_repr, f_eval) in tag_functions]

    pass


class MakeSymbolicLinks(odf_file.Command):

    """Execute the *ln* shell command.

    """

    def __init__(self, old, new):
        args = 'ln -s "%s" "%s"' % (old, new)
        STD_ERROR.put('%s\n' % args)
        odf_file.Command.__init__(
            self,
            args=args,
            )
        return

    def run(self):
        self.wait()
        return self


class MakeGuids(odf_file.Command):

    """Execute the *gnucash-make_guids* shell command.

    New guids are required when adding xml objects to a GnuCash book.

    """

    def __init__(self, count=1):
        odf_file.Command.__init__(
            self,
            args='gnucash-make-guids %i' % count,
            stdout=odf_file.COMMAND_PIPE,
            stderr=odf_file.COMMAND_STDOUT,
            )
        return

    def run(self):

        """Return a list of GUIDs.

        """
        
        self.wait()
        result = self.stdout.read().split()
        return result


class GncObject(object):

    """Abstract GnuCash object.

    """

    uri = {
        'gnc': "http://www.gnucash.org/XML/gnc",
        'act': "http://www.gnucash.org/XML/act",
        'book': "http://www.gnucash.org/XML/book",
        'cd': "http://www.gnucash.org/XML/cd",
        'cmdty': "http://www.gnucash.org/XML/cmdty",
        'price': "http://www.gnucash.org/XML/price",
        'slot': "http://www.gnucash.org/XML/slot",
        'split': "http://www.gnucash.org/XML/split",
        'sx': "http://www.gnucash.org/XML/sx",
        'trn': "http://www.gnucash.org/XML/trn",
        'ts': "http://www.gnucash.org/XML/ts",
        'fs': "http://www.gnucash.org/XML/fs",
        'bgt': "http://www.gnucash.org/XML/bgt",
        'recurrence': "http://www.gnucash.org/XML/recurrence",
        'lot': "http://www.gnucash.org/XML/lot",
        'cust': "http://www.gnucash.org/XML/cust",
        'job': "http://www.gnucash.org/XML/job",
        'addr': "http://www.gnucash.org/XML/addr",
        'owner': "http://www.gnucash.org/XML/owner",
        'taxtable': "http://www.gnucash.org/XML/taxtable",
        'tte': "http://www.gnucash.org/XML/tte",
        'employee': "http://www.gnucash.org/XML/employee",
        'order': "http://www.gnucash.org/XML/order",
        'billterm': "http://www.gnucash.org/XML/billterm",
        'bt-days': "http://www.gnucash.org/XML/bt-days",
        'bt-prox': "http://www.gnucash.org/XML/bt-prox",
        'invoice': "http://www.gnucash.org/XML/invoice",
        'entry': "http://www.gnucash.org/XML/entry",
        'vendor': "http://www.gnucash.org/XML/vendor",
        }

    def from_elt(self, tree=None):

        """Abstract.

        """
        
        self.tree = tree
        return self
     
    def to_elt(self, tree=None):

        """Abstract.

        """
        
        self.tree = tree
        return self

    def ns(self, path):

        """Expand namespace prefixes.
        
        """
        
        tags = path.split('/')
        for (ndx, tag) in enumerate(tags):
            qfn = tag.split(':')
            if len(qfn) > 1:
                (prefix, tag) = qfn[:2]
                prefix = self.uri[prefix]
                tag = '{%s}%s' % (prefix, tag)
                tags[ndx] = tag
        path = '/'.join(tags)
        return path
     
    def find_iter(self, path):
        return self.tree.iterfind(self.ns(path))

    def find_1(self, path):
        return self.tree.find(self.ns(path))

    def get(self, path):
        result = self.find_1(path)
        if result is None:
            return None
        return result.text

    def put(self, path, text, can_add=False):
        elt = self.find_1(path)
        if elt is None:
            if can_add:
                if text in EMPTY:
                    pass
                else:
                    elt = ET.SubElement(
                        self.tree,
                        self.ns(path),
                        )
                    elt.text = text
            else:
                raise NotImplementedError
        else:
            elt.text = text
        return self

    def get_date(self, path):
        text = self.get(path)
        if text is None:
            return None
        result = self.date_from_string(text)
        return result

    def date_from_string(self, text):
        match = PAT_GNC_DATE.match(text)
        if match is None:
            raise NotImplementedError
        tokens = [int(tok, 10) for tok in match.group(1,2,3,4,5,6,7,8)]
        result = GncDateTime(*tokens[:6])
        (delta_hrs, delta_min) = tokens[6:]
        if delta_hrs < ZERO:
            delta_min = - delta_min
        result.set_delta(delta_hrs * 60 + delta_min)
        return result

    def put_date(self, path, date):
        return self.put(path, str(date))

    def get_dollars(self, path):
        text = self.get(path)
        if text is None:
            result = None
        else:
            result = self.numeric_from_string(text)
        return result

    def numeric_from_string(self, text):
        (numerator, denominator) = text.split('/')
        if denominator in ['100']:
            pass
        else:
            raise NotImplementedError
        result = float(numerator) / 100.0
        return result

    def put_dollars(self, path, value):
        self.put(path, self.numeric_to_string(value))
        return self

    def numeric_to_string(self, value):
        result = '%i/100' % int(value * 100)
        return result


class GncSlot(GncObject):

    """An xml key/value object.

    GnuCash books use 'slots' in various objects at several levels of
    heirarchy to record object attributes.

    """

    def append_to_slots(self, slots, key, value, _type='string'):
        slot = ET.SubElement(
            slots,
            'slot',
            )
        ET.SubElement(
            slot,
            self.ns('slot:key'),
            ).text = key
        if _type in ['numeric']:
            value = self.numeric_to_string(value)
        ET.SubElement(
            slot,
            self.ns('slot:value'),
            attrib={'type': _type},
            ).text = value
        return self


class GncSplit(GncObject):

    """GnuCash transaction split.

    This holds one positive or negative value, a debit or credit.  A
    transaction generally holds at least two 'splits.'  All the debit
    splits must balance all the credit splits in each transaction and
    thus in the whole book.

    """

    def from_elt(self, tree):
        GncObject.from_elt(self, tree)
        self.uid = self.get('split:id')
        self.reconciled_state = self.get('split:reconciled-state')
        self.value = self.get_dollars('split:value')
        self.quantity = self.get_dollars('split:quantity')
        self.memo = self.get('split:memo')
        self.account_uid = self.get('split:account')
        return self

    def is_credit(self):
        return self.value < ZERO

    def is_reconciled(self):
        return self.reconciled_state not in ['N', 'n']

    def append_to_splits(
        self,
        splits,
        reconciled_state,
        value,
        quantity,
        account_uid,
        memo=None,
        slots=None,
        ):
        self.uid = MakeGuids().run()[ZERO]
        split = ET.SubElement(
            splits,
            self.ns('trn:split'),
            )
        ET.SubElement(
            split,
            self.ns('split:id'),
            attrib={'type': 'guid'},
            ).text = self.uid
        ET.SubElement(
            split,
            self.ns('split:action'),
            ).text = 'check'
        ET.SubElement(
            split,
            self.ns('split:reconciled-state'),
            ).text = reconciled_state
        ET.SubElement(
            split,
            self.ns('split:value'),
            ).text = self.numeric_to_string(value)
        ET.SubElement(
            split,
            self.ns('split:quantity'),
            ).text = self.numeric_to_string(quantity)
        ET.SubElement(
            split,
            self.ns('split:account'),
            attrib={'type': 'guid'},
            ).text = account_uid
        if memo is None:
            pass
        else:
            ET.SubElement(
                split,
                self.ns('split:memo'),
                ).text = memo
        if slots is None:
            pass
        else:
            slot_s = ET.SubElement(
                split,
                self.ns('split:slots'),
                )
            slot = GncSlot()
            for (key, value, _type) in slots:
                slot.append_to_slots(slot_s, key, value, _type)
        return self


class GncTransaction(GncObject):

    """GnuCash transaction.

    """

    def from_elt(self, tree):
        GncObject.from_elt(self, tree)
        self.uid = self.get('trn:id')
        self.date_posted = self.get_date('trn:date-posted/ts:date')
        self.description = self.get('trn:description')
        self.form_num = self.get('trn:num')
        self.splits = [
            GncSplit().from_elt(elt)
            for elt in self.find_iter('trn:splits/trn:split')
            ]
        return self

    def is_marked(self):

        """You tell Pretty Checks which bank-account credits are ready
        to be printed by changing their form numbers to a single
        asterisk.  You provide the payee name in the transaction
        descrtipion.

        """
        
        return (not self.description in EMPTY) and (self.form_num in ['*'])


class GncVoid(GncTransaction):

    """GnuCash voided form.

    """

    def append_to_book(
        self,
        book,
        account,
        form_num,
        old_amount=ZERO,
        old_value=ZERO,
        void_date=None,
        ):
        self.uid = MakeGuids().run()[ZERO]
        if void_date is None:
            void_date = TODAY
        trn = ET.SubElement(
            book.main_elt,
            self.ns('gnc:transaction'),
            version=GNC_VERSION,
            )
        ET.SubElement(
            trn,
            self.ns('trn:id'),
            attrib={'type': 'guid'},
            ).text = self.uid
        self.append_currency(trn)
        ET.SubElement(
            trn,
            self.ns('trn:num'),
            ).text = form_num
        self.append_date(trn, 'trn:date-posted', void_date)
        self.append_date(trn, 'trn:date-entered', NOW)
        ET.SubElement(
            trn,
            self.ns('trn:description'),
            ).text = 'Void Check'
        slots = ET.SubElement(
            trn,
            self.ns('trn:slots'),
            )
        slot = GncSlot()
        for (key, value, _type) in [
            ('notes', 'Voided transaction', 'string'),
            ('void-reason', 'Scratch check form', 'string'),
            ('void-time', str(NOW), 'string'),
            ('trans-read-only', 'Transaction Voided', 'string'),
            ]:
            slot.append_to_slots(slots, key, value, _type)
        splits = ET.SubElement(
            trn,
            self.ns('trn:splits'),
            )
        split = GncSplit()
        split.append_to_splits(
            splits,
            reconciled_state='v',
            value=ZERO,
            quantity=ZERO,
            account_uid=account.uid,
            slots=[
              ('void-former-amount', old_amount, 'numeric'),
              ('void-former-value', old_value, 'numeric'),
              ],
            )
        book.count_transactions += 1
        return self

    def append_currency(self, trn):
        currency = ET.SubElement(
            trn,
            self.ns('trn:currency'),
            )
        ET.SubElement(
            currency,
            self.ns('cmdty:space'),
            ).text = 'ISO4217'
        ET.SubElement(
            currency,
            self.ns('cmdty:id'),
            ).text = 'USD'
        return self

    def append_date(self, trn, tag, text):
        date = ET.SubElement(
            trn,
            self.ns(tag),
            )
        ET.SubElement(
            date,
            self.ns('ts:date'),
            ).text = str(text)
        return self


class GncVendorAddress(GncObject):

    """GnuCash vendor address.

    """

    def from_elt(self, tree):
        GncObject.from_elt(self, tree)
        self.addr_name = self.get('addr:name')
        self.addr_1 = self.get('addr:addr1')
        self.addr_2 = self.get('addr:addr2')
        self.addr_3 = self.get('addr:addr3')
        self.addr_4 = self.get('addr:addr4')
        self.phone = self.get('addr:phone')
        self.fax = self.get('addr:fax')
        self.email = self.get('addr:email')
        return self

    def to_elt(self, tree):
        GncObject.to_elt(self, tree)
        self.put('addr:name', self.addr_name)
        self.put('addr:addr1', self.addr_1, can_add=True)
        self.put('addr:addr2', self.addr_2, can_add=True)
        self.put('addr:addr3', self.addr_3, can_add=True)
        self.put('addr:addr4', self.addr_4, can_add=True)
        self.put('addr:phone', self.phone, can_add=True)
        self.put('addr:fax', self.fax, can_add=True)
        self.put('addr:email', self.email, can_add=True)
        return self

    def append_to_vendor(self, vendor):
        addr = ET.SubElement(
            vendor,
            self.ns('vendor:addr'),
            version=GNC_VERSION,
            )
        ET.SubElement(
            addr,
            self.ns('addr:name'),
            ).text = self.addr_name
        for (tag, value) in [
            ('addr:addr1', self.addr_1),
            ('addr:addr2', self.addr_2),
            ('addr:addr3', self.addr_3),
            ('addr:addr4', self.addr_4),
            ('addr:phone', self.phone),
            ('addr:fax', self.fax),
            ('addr:email', self.email),
            ]:
            if value in EMPTY:
                pass
            else:
                ET.SubElement(
                    addr,
                    self.ns(tag),
                    ).text = value
        return self


class GncVendor(GncObject):

    """GnuCash vendor.

    """

    def from_elt(self, tree):
        GncObject.from_elt(self, tree)
        self.uid = self.get('vendor:guid')
        self.name = self.get('vendor:name')
        self.id = self.get('vendor:id')
        self.notes = self.get('vendor:notes')
        elt = self.find_1('vendor:addr')
        if elt is None:
            self.address = None
        else:
            self.address = GncVendorAddress().from_elt(elt)
        return self

    def to_elt(self, tree):
        GncObject.to_elt(self, tree)
        self.put('vendor:name', self.name)
        self.put('vendor:id', self.id)
        self.put('vendor:notes', self.notes, can_add=True)
        elt = self.find_1('vendor:addr')
        if elt is None:
            raise NotImplementedError
        if self.address is None:
            raise NotImplementedError
        self.address.to_elt(elt)
        return self

    def append_to_book(self, book):
        self.uid = MakeGuids().run()[ZERO]
        vendor = ET.SubElement(
            book.main_elt,
            self.ns('gnc:GncVendor'),
            version=GNC_VERSION,
            )
        ET.SubElement(
            vendor,
            self.ns('vendor:guid'),
            attrib={'type': 'guid'},
            ).text = self.uid
        ET.SubElement(
            vendor,
            self.ns('vendor:name'),
            ).text = self.name
        ET.SubElement(
            vendor,
            self.ns('vendor:id'),
            ).text = self.id
        if self.address is None:
            pass
        else:
            self.address.append_to_vendor(vendor)
        ET.SubElement(
            vendor,
            self.ns('vendor:notes'),
            ).text = self.notes
        ET.SubElement(
            vendor,
            self.ns('vendor:taxincluded'),
            ).text = 'USEGLOBAL'
        ET.SubElement(
            vendor,
            self.ns('vendor:active'),
            ).text = '1'
        currency = ET.SubElement(
            vendor,
            self.ns('vendor:currency'),
            )
        ET.SubElement(
            currency,
            self.ns('cmdty:space'),
            ).text = 'ISO4217'
        ET.SubElement(
            currency,
            self.ns('cmdty:id'),
            ).text = 'USD'
        ET.SubElement(
            vendor,
            self.ns('vendor:use-tt'),
            ).text = '0'
        book.count_vendors += 1
        return self


class GncAccount(GncObject):

    """GnuCash account.

    """

    def from_elt(self, tree):
        GncObject.from_elt(self, tree)
        self.uid = self.get('act:id')
        self.name = self.get('act:name')
        self.type = self.get('act:type')
        self.description = self.get('act:description')
        return self

    def is_bank(self):
        return self.type in ['BANK']


class GncBook(GncObject):

    """GnuCash book.

    """

    def load(self, file_name):
        self.tree = ET.ElementTree()
        try:
            flob = gzip.open(file_name, 'rb')
            self.tree.parse(flob)
            ok = True
        except IOError:
            ok = False
        if not ok:
            try:
                flob = open(file_name, 'rb')
                self.tree.parse(flob)
                ok = True
            except IOError:
                ok = False
        if not ok:
            raise NotImplementedError
        else:
            flob.close()
        self.assert_version()
        self.get_counts()
        return self

    def save(self, file_name, zip=False):
        self.put_counts()
        if zip:
            flob = gzip.open(file_name, 'wb')
        else:
            flob = open(file_name, 'wb')
        self.tree.write(flob, encoding=CODING, xml_declaration=True, pretty_print=True)
        flob.close()
        return self

    def find_all_transactions(self):

        """Raise AssertionError for damaged books.

        """
        
        self.transactions = [
            GncTransaction().from_elt(elt)
            for elt in self.find_iter('.//gnc:transaction')
            ]
        assert len(self.transactions) == self.count_transactions
        return self

    def find_all_vendors(self):

        """Raise AssertionError for damaged books.

        """
        
        self.vendors = [
            GncVendor().from_elt(elt)
            for elt in self.find_iter('.//gnc:GncVendor')
            ]
        assert len(self.vendors) == self.count_vendors
        return self

    def find_all_accounts(self):

        """Raise AssertionError for damaged books.

        """
        
        self.accounts = [
            GncAccount().from_elt(elt)
            for elt in self.find_iter('.//gnc:account')
            ]
        assert len(self.accounts) == self.count_accounts
        return self

    def find_main_elt(self):

        """Raise AssertionError for empty books.

        """
        
        self.main_elt = self.find_1('gnc:book')
        assert not self.main_elt is None
        return self

    def get_version(self):
        self.find_main_elt()
        result = self.main_elt.attrib.get('version')
        return result

    def assert_version(self, version=GNC_VERSION):

        """Raise AssertionError for incompatible versions GnuCash book
        files.

        """
        
        assert self.get_version() == version
        return self

    def get_counts(self):
        counts = self.find_iter('.//gnc:count-data')
        key = self.ns('cd:type')
        for count in counts:
            if count.attrib[key] in ['account']:
                self.count_accounts = int(count.text, 10)
            elif count.attrib[key] in ['transaction']:
                self.count_transactions = int(count.text, 10)
            elif count.attrib[key] in ['gnc:GncVendor']:
                self.count_vendors = int(count.text, 10)
        return self

    def put_counts(self):
        counts = self.find_iter('.//gnc:count-data')
        key = self.ns('cd:type')
        for count in counts:
            if count.attrib[key] in ['account']:
                count.text = str(self.count_accounts)
            elif count.attrib[key] in ['transaction']:
                count.text = str(self.count_transactions)
            elif count.attrib[key] in ['gnc:GncVendor']:
                count.text = str(self.count_vendors)
        return self

    def find_marked_checks(self, acct):
        result = []
        for transaction in self.transactions:
            if transaction.is_marked():
                for split in transaction.splits:
                    if (split.account_uid == acct.uid) and split.is_credit():
                        result.append([transaction, split])
        return result

    def find_vendor(self, **tags):
        for vendor in self.vendors:
            for (tag, value) in tags.iteritems():
                if getattr(vendor, tag) == value:
                    pass
                else:
                    break
            else:
                return vendor
        else:
            return None

    def bump(self, num, descending=False):

        """Increment/decrement check numbers.

        Check numbers must have a numeric part but need not be wholly
        numeric.  The first numeric part is incremented.

        """
        
        if descending:
            inc = -1
        else:
            inc = 1
        match = PAT_NUMERIC.search(num)
        if match is None:
            result = num
        else:
            seq = match.group(1)
            digits = len(seq)
            seq = int(seq, 10)
            seq += inc
            seq = '0'*digits + str(seq)
            seq = seq[-digits:]
            result = num[:match.start()] + seq + num[match.end(1):]
        return result


class PcScript(optparse.OptionParser):

    """Abstract script.

    For notes on descendant classes see DOC_STRINGS.

    """

    def __init__(self, usage=None, **args):
        if usage is None:
            usage = """usage:  %%prog [--setup] [--readme]
        
%s""" % DOC_STRINGS[self.script_name].strip()
        optparse.OptionParser.__init__(self, usage=usage, **args)
        return

    def run(self):
        (self.opts, self.args) = self.parse_args()
        return self


class Pcs(PcScript):

    script_name = 'pcs'
    
    def __init__(self, usage=None, **args):
        usage = """usage:  %%prog [--setup] [--readme]
        
%s""" % DOC_STRINGS[self.script_name].strip()
        PcScript.__init__(self, usage=usage, **args)
        self.add_option('-i', '--setup',
                        help=
                        
                        """Create necessary symbolic links to this
                        script.  Run this once after unpacking the
                        script source.
                        
                        """,
                        
                        action='store_true',
                        default=False,
                        )
        self.add_option('-r', '--readme',
                        help="Generate the README file for Pretty Checks.",
                        action='store_true',
                        default=False,
                        )
        return

    def dump_doc_strings(self):
        description = [__doc__]
        for script_name in ['pcs'] + SCRIPT_NAMES:
            h2 = 'o %s' % script_name
            description.append('\n%s\n' % h2)
            description.append('=' * len(h2))
            description.append(DOC_STRINGS[script_name])
        STD_ERROR.put(NULL.join(description))
        return self

    def run(self):
        result = PcScript.run(self)
        if self.opts.setup:
            for script_name in SCRIPT_NAMES:
                MakeSymbolicLinks(old=sys.argv[ZERO], new='%s.py' % script_name).run()
        if self.opts.readme:
            self.dump_doc_strings()
        return result


class PcsExtractVendors(PcScript):

    script_name = 'pcs_extract_vendors'
    
    def __init__(self):
        usage = """usage:  %%prog --book=BOOK [CSV]
        
%s""" % DOC_STRINGS[self.script_name].strip()
        PcScript.__init__(self, usage=usage)
        self.add_option('-b', '--book',
                        help='GnuCash book.  (A file name is required.)',
                        )
        return

    def run(self):
        result = PcScript.run(self)
        if self.opts.book is None:
            self.error('--book option is required')
        if os.path.exists(self.opts.book):
            pass
        else:
            self.error('--book file "%s" does not exist' % self.opts.book)
        book = GncBook().load(self.opts.book)
        book.find_all_vendors()
        csv = CsvVendors()
        for vendor in book.vendors:
            rec = [(tag, getattr(vendor, tag)) for tag in csv.tags_id]
            if vendor.address is None:
                pass
            else:
                rec.extend([(tag, getattr(vendor.address, tag)) for tag in csv.tags_addr])
            csv.append(dict(rec))
        if self.args:
            csv.save(self.args[ZERO])
        else:
            csv.save()
        STD_ERROR.put('%i vendors extracted.\n' % len(csv))
        return result


class PcsMergeVendors(PcScript):

    script_name = 'pcs_merge_vendors'
    
    def __init__(self):
        usage = """usage:  %%prog --book=BOOK [CSV]
        
%s""" % DOC_STRINGS[self.script_name].strip()
        PcScript.__init__(self, usage=usage)
        self.add_option('-b', '--book',
                        help='GnuCash book.  (A file name is required.)',
                        )
        return

    def run(self):
        result = PcScript.run(self)
        if self.opts.book is None:
            self.error('--book option is required')
        if os.path.exists(self.opts.book):
            pass
        else:
            self.error('--book file "%s" does not exist' % self.opts.book)
        book = GncBook().load(self.opts.book)
        book.find_all_vendors()
        csv = CsvVendors()
        if self.args:
            csv.load(self.args[ZERO])
        else:
            csv.load()
        for rec in csv:
            vendor = GncVendor()
            vendor.address = GncVendorAddress()
            for tag in csv.tags_id:
                setattr(vendor, tag, rec.get(tag))
            for tag in csv.tags_addr:
                setattr(vendor.address, tag, rec.get(tag))
            target = book.find_vendor(uid=vendor.uid)
            if target is None:
                vendor.append_to_book(book)
            else:
                vendor.to_elt(target.tree)
        book.save(self.opts.book)
        STD_ERROR.put('%i vendors merged.\n' % len(csv))
        return result


class PcsExtractPayments(PcScript):

    script_name = 'pcs_extract_payments'
    
    def __init__(self):
        usage = """usage:  %%prog --book=BOOK --acct=ACCT [CSV]
        
%s""" % DOC_STRINGS[self.script_name].strip()
        PcScript.__init__(self, usage=usage)
        self.add_option('-b', '--book',
                        help='GnuCash book.  (A file name is required.)',
                        )
        self.add_option('-a', '--acct',
                        help='Account "Name"  (Use double quotes.)',
                        )
        return

    def run(self):
        result = PcScript.run(self)
        if self.opts.book is None:
            self.error('--book option is required')
        if os.path.exists(self.opts.book):
            pass
        else:
            self.error('--book file "%s" does not exist' % self.opts.book)
        book = GncBook().load(self.opts.book)
        book.find_all_transactions()
        book.find_all_accounts()
        book.bank_accounts = [acct for acct in book.accounts if acct.is_bank()]
        for acct in book.bank_accounts:
            if acct.name == self.opts.acct:
                break
        else:
            diagnostic = ['--acct name not found']
            diagnostic.append('Valid bank accounts are:')
            diagnostic.extend(['  %s' % acct.name for acct in book.bank_accounts])
            self.error('\n'.join(diagnostic))
        csv = CsvPayments()
        for (transaction, split) in book.find_marked_checks(acct):
            check = Check().from_gnc(transaction, split)
            rec = {}
            for (tag, f_repr, f_eval) in csv.tag_functions:
                if f_repr is None:
                    rec[tag] = getattr(check, tag)
                elif isinstance(f_repr, str):
                    rec[tag] = f_repr % getattr(check, tag)
                else:
                    rec[tag] = f_repr(getattr(check, tag))
            csv.append(rec)
        if self.args:
            csv.save(self.args[ZERO])
        else:
            csv.save()
        STD_ERROR.put('%i payments extracted.\n' % len(csv))
        return result


class PcsMergePayments(PcScript):

    script_name = 'pcs_merge_payments'
    
    def __init__(self):
        usage = """usage:  %%prog --book=BOOK [CSV]
        
%s""" % DOC_STRINGS[self.script_name].strip()
        PcScript.__init__(self, usage=usage)
        self.add_option('-b', '--book',
                        help='GnuCash book.  (A file name is required.)',
                        )
        return

    def run(self):
        result = PcScript.run(self)
        if self.opts.book is None:
            self.error('--book option is required')
        if os.path.exists(self.opts.book):
            pass
        else:
            self.error('--book file "%s" does not exist' % self.opts.book)
        book = GncBook().load(self.opts.book)
        csv = CsvPayments()
        if self.args:
            csv.load(self.args[ZERO])
        else:
            csv.load()
        splits = {}
        for rec in csv:
            check = Check()
            for (tag, f_repr, f_eval) in csv.tag_functions:
                if f_eval is None:
                    setattr(check, tag, rec[tag])
                else:
                    setattr(check, tag, f_eval(rec[tag]))
            splits[check.split_uid] = check
        count = ZERO
        for split in book.find_iter('.//trn:split'):
            uid = split.find(book.ns('split:id')).text
            if uid is None:
                pass
            else:
                if uid in splits:
                    check = splits[uid]
                    check.to_gnc(split)
                    count += 1
        book.save(self.opts.book)
        STD_ERROR.put('%i payments merged.\n' % count)
        return result


class PcsReprintChecks(PcScript):

    script_name = 'pcs_reprint_checks'
    
    def __init__(self):
        usage = """usage:  %%prog --book=BOOK --start=START [--descending]
          [--template=ODT] --checks=ODF [--postdate=YYYYMMDD] [CSV]
        
%s""" % DOC_STRINGS[self.script_name].strip()
        PcScript.__init__(self, usage=usage)
        self.add_option('-b', '--book',
                        help='GnuCash book.  (A file name is required.)',
                        )
        self.add_option('-s', '--start',
                        help='First check form number to print.',
                        )
        self.add_option('-d', '--descending',
                        help=

                        """Decrement rather than increment from START.
                        This may be helpful for printers that pull
                        from the top of a face-down stack of blanks in
                        descending order.  Be sure to count out the
                        exact number of forms required before stacking
                        them to avoid having surplus, out-of-sequence
                        forms left over.  Alternatively restack blanks
                        face down in ascending order.  Default is
                        ascending.""",

                        action='store_true',
                        default=False,
                        )
        self.add_option('-t', '--template',
                        help='Check-form skeleton.  (Default is check_form.odt)',
                        default='check_form.odt',
                        metavar='ODT',
                        )
        self.add_option('-c', '--checks',
                        help=

                        """Output checks.  (A valid file name is
                        required, but the file need not necessarily
                        exist.  It will be created.)""",
                        
                        metavar='ODF',
                        )
        self.add_option('-p', '--postdate',
                        help='Posting date.  (Default is %s.)' % TODAY.strftime('%Y%m%d'),
                        default=TODAY.strftime('%Y%m%d'),
                        metavar='YYYYMMDD',
                        )
        return

    def run(self):
        result = PcScript.run(self)
        if self.opts.book is None:
            self.error('--book option is required')
        if os.path.exists(self.opts.book):
            pass
        else:
            self.error('--book file "%s" does not exist' % self.opts.book)
        if os.path.exists(self.opts.template):
            pass
        else:
            self.error('--template file "%s" does not exist' % self.opts.template)
        if self.opts.start:
            form_num = self.opts.start
        else:
            self.error('--start check form number is required')
        if self.opts.checks:
            pass
        else:
            self.error('--checks output file name is required')
        try:
            year = int(self.opts.postdate[ZERO:4], 10)
            month = int(self.opts.postdate[4:6], 10)
            day = int(self.opts.postdate[6:8], 10)
            date_posted = GncDateTime(year, month, day).set_delta(TZ_OFFSET_MINUTES)
        except (ValueError):
            self.error('--postdate is invalid')
        book = GncBook().load(self.opts.book)
        book.find_all_vendors()
        content = odf_file.Content()
        skeleton = content.load_soup(flob=self.opts.template)
        frame = skeleton.find(odf_file.ns('.//text:section'))
        page = odf_file.Template().extract(frame)
        csv = CsvPayments()
        if self.args:
            csv.load(self.args[ZERO])
        else:
            csv.load()
        for rec in csv:
            check = Check()
            for (tag, f_repr, f_eval) in csv.tag_functions:
                if f_eval is None:
                    setattr(check, tag, rec[tag])
                else:
                    setattr(check, tag, f_eval(rec[tag]))
            vendor = book.find_vendor(name=check.vendor_name)
            check.from_vendor(vendor)
            check.form_num = form_num
            form_num = book.bump(form_num, descending=self.opts.descending)
            rec['form_num'] = check.form_num
            check.date_posted = date_posted
            rec['date_posted'] = repr(check.date_posted)

            # Raise AssertionError for checks without payee.

            assert not check.payee in EMPTY

            # Raise AssertionError for checks without positive value.            
            
            assert check.value > ZERO
            check.to_odf(page)
        if self.args:
            csv.save(self.args[ZERO])
        else:
            csv.save()
        content.save_soup(ET.tostring(skeleton), flob=self.opts.checks)
        STD_ERROR.put('%i checks to be printed.\n' % len(csv))
        return result


class PcsVoidChecks(PcScript):

    script_name = 'pcs_void_checks'
    
    def __init__(self):
        usage = """usage:  --book=BOOK --acct=ACCT --start=START --end=END
        
%s""" % DOC_STRINGS[self.script_name].strip()
        PcScript.__init__(self, usage=usage)
        self.add_option('-b', '--book',
                        help='GnuCash book.  (A file name is required.)',
                        )
        self.add_option('-a', '--acct',
                        help='Account "Name"  (Use double quotes.)',
                        )
        self.add_option('-s', '--start',
                        help='Low beginning check form number to void.',
                        )
        self.add_option('-e', '--end',
                        help='High stopping check form number to void.',
                        )
        return

    def run(self):
        result = PcScript.run(self)
        if self.opts.book is None:
            self.error('--book option is required')
        if os.path.exists(self.opts.book):
            pass
        else:
            self.error('--book file "%s" does not exist' % self.opts.book)
        book = GncBook().load(self.opts.book)
        book.find_all_accounts()
        book.bank_accounts = [acct for acct in book.accounts if acct.is_bank()]
        for acct in book.bank_accounts:
            if acct.name == self.opts.acct:
                break
        else:
            diagnostic = ['--acct name not found']
            diagnostic.append('Valid bank accounts are:')
            diagnostic.extend(['  %s' % acct.name for acct in book.bank_accounts])
            self.error('\n'.join(diagnostic))
        num = self.opts.start
        forms = [num]
        while True:
            num = book.bump(num)
            if num > self.opts.end:
                break
            forms.append(num)
        count_forms = len(forms)
        if count_forms > 20:
            try:
                veto = raw_input('You are about to void %i check forms.  OK?  (n/y)  ' %
                                 count_forms)
            except EOFError:
                veto = 'n'
        else:
            veto = 'y'
        if veto in ['y', 'Y']:
            pass
        else:
            self.error('trying to void too many forms')
        void = GncVoid()
        for num in forms:
            void.append_to_book(
                book,
                account=acct,
                form_num=num,
                )
        book.save(self.opts.book)
        STD_ERROR.put('%i check forms voided.\n' % count_forms)
        return result


if __name__ == "__main__":

    # Skip this if pcs.py is imported as a module.
    
    if DEBUG:
        import doctest
        doctest.testmod()
    (SCRIPT_NAME, EXT) = os.path.splitext(os.path.basename(sys.argv[ZERO]))
    if SCRIPT_NAME in ['pcs_extract_vendors']:
        PcsExtractVendors().run()
    elif SCRIPT_NAME in ['pcs_merge_vendors']:
        PcsMergeVendors().run()
    elif SCRIPT_NAME in ['pcs_extract_payments']:
        PcsExtractPayments().run()
    elif SCRIPT_NAME in ['pcs_reprint_checks']:
        PcsReprintChecks().run()
    elif SCRIPT_NAME in ['pcs_void_checks']:
        PcsVoidChecks().run()
    elif SCRIPT_NAME in ['pcs_merge_payments']:
        PcsMergePayments().run()
    else:
        Pcs().run()

# Fin
